@@ -288,7 +288,6 @@ static rpmRC handleOneTrigger(rpmts ts, rpmte te, rpmsenseFlags sense,
 	return rc;
 
     headerGet(trigH, RPMTAG_INSTPREFIXES, &pfx, HEADERGET_ALLOC|HEADERGET_ARGV);
-    (void) rpmdsSetNoPromote(trigger, 1);
 
     while ((i = rpmdsNext(trigger)) >= 0) {
 	uint32_t tix;
@@ -436,6 +435,21 @@ static rpmpsm rpmpsmFree(rpmpsm psm)
     return NULL;
 }
 
+static int isUpdate(rpmts ts, rpmte te)
+{
+    rpmtsi pi = rpmtsiInit(ts);
+    rpmte p;
+    int update = 0;
+    while ((p = rpmtsiNext(pi, TR_REMOVED)) != NULL) {
+	if (rpmteDependsOn(p) == te) {
+	    update = 1;
+	    break;
+	}
+    }
+    rpmtsiFree(pi);
+    return update;
+}
+
 static rpmpsm rpmpsmNew(rpmts ts, rpmte te, pkgGoal goal)
 {
     rpmpsm psm = xcalloc(1, sizeof(*psm));
@@ -461,7 +475,7 @@ static rpmpsm rpmpsmNew(rpmts ts, rpmte te, pkgGoal goal)
 	    break;
 	case PKG_VERIFY:
 	case PKG_POSTTRANS:
-	    psm->scriptArg = npkgs_installed;
+	    psm->scriptArg = npkgs_installed + isUpdate(psm->ts, psm->te);
 	    psm->countCorrection = 0;
 	    break;
 	default:
@@ -536,7 +550,7 @@ static void markReplacedInstance(rpmts ts, rpmte te)
 static rpmRC dbAdd(rpmts ts, rpmte te)
 {
     Header h = rpmteHeader(te);
-    rpm_time_t installTime = (rpm_time_t) time(NULL);
+    rpm_time_t installTime = rpmtsGetTime(ts, 1);
     rpmfs fs = rpmteGetFileStates(te);
     rpm_count_t fc = rpmfsFC(fs);
     rpm_fstate_t * fileStates = rpmfsGetStates(fs);
@@ -672,21 +686,26 @@ static rpmRC rpmPackageInstall(rpmts ts, rpmpsm psm)
 	    if (rc) break;
 	}
 
-	rc = rpmpsmUnpack(psm);
+	if ((rc = rpmChrootIn()) == 0) {
+	    rc = rpmpsmUnpack(psm);
+	    rpmChrootOut();
+	}
 	if (rc) break;
 
-	/*
-	 * If this package has already been installed, remove it from
-	 * the database before adding the new one.
-	 */
-	if (rpmteDBInstance(psm->te)) {
-	    rc = dbRemove(ts, psm->te);
+	if (!(rpmtsFlags(ts) & RPMTRANS_FLAG_NODB)) {
+	    /*
+	     * If this package has already been installed, remove it from
+	     * the database before adding the new one.
+	     */
+	    if (rpmteDBInstance(psm->te)) {
+		rc = dbRemove(ts, psm->te);
+		if (rc) break;
+	    }
+
+	    rc = dbAdd(ts, psm->te);
 	    if (rc) break;
 	}
 
-	rc = dbAdd(ts, psm->te);
-	if (rc) break;
-
 	if (!(rpmtsFlags(ts) & RPMTRANS_FLAG_NOTRIGGERIN)) {
 	    /* Run upper file triggers i. e. with higher priorities */
 	    /* Run file triggers in other package(s) this package sets off. */
@@ -779,7 +798,10 @@ static rpmRC rpmPackageErase(rpmts ts, rpmpsm psm)
 	    if (rc) break;
 	}
 
-	rc = rpmpsmRemove(psm);
+	if ((rc = rpmChrootIn()) == 0) {
+	    rc = rpmpsmRemove(psm);
+	    rpmChrootOut();
+	}
 	if (rc) break;
 
 	/* Run file triggers in other package(s) this package sets off. */
@@ -809,7 +831,8 @@ static rpmRC rpmPackageErase(rpmts ts, rpmpsm psm)
 	    rpmtriggersPrepPostUnTransFileTrigs(psm->ts, psm->te);
 	}
 
-	rc = dbRemove(ts, psm->te);
+	if (!(rpmtsFlags(ts) & RPMTRANS_FLAG_NODB))
+	    rc = dbRemove(ts, psm->te);
     }
 
     rpmswExit(rpmtsOp(psm->ts, RPMTS_OP_ERASE), 0);
@@ -817,11 +840,43 @@ static rpmRC rpmPackageErase(rpmts ts, rpmpsm psm)
     return rc;
 }
 
+static rpmRC rpmPackageRestore(rpmts ts, rpmpsm psm)
+{
+    rpmRC rc = RPMRC_OK;
+
+    rpmswEnter(rpmtsOp(psm->ts, RPMTS_OP_INSTALL), 0);
+    if ((rc = rpmChrootIn()) == 0) {
+	char *failedFile = NULL;
+	rpmpsmNotify(psm, RPMCALLBACK_INST_START, 0);
+	/* make sure first progress call gets made */
+	rpmpsmNotify(psm, RPMCALLBACK_INST_PROGRESS, 0);
+
+	if (!(rpmtsFlags(psm->ts) & RPMTRANS_FLAG_JUSTDB)) {
+	    if (rpmfilesFC(psm->files) > 0) {
+		int fsmrc = rpmPackageFilesInstall(psm->ts, psm->te, psm->files,
+					      psm, &failedFile);
+		rc = (fsmrc == 0) ? RPMRC_OK : RPMRC_FAIL;
+	    }
+	}
+
+	/* XXX make sure progress reaches 100% */
+	rpmpsmNotify(psm, RPMCALLBACK_INST_PROGRESS, psm->total);
+	rpmpsmNotify(psm, RPMCALLBACK_INST_STOP, psm->total);
+
+	free(failedFile);
+	rpmChrootOut();
+    }
+    rpmswExit(rpmtsOp(psm->ts, RPMTS_OP_INSTALL), 0);
+
+    return rc;
+}
+
 static const char * pkgGoalString(pkgGoal goal)
 {
     switch (goal) {
     case PKG_INSTALL:	return "  install";
     case PKG_ERASE:	return "    erase";
+    case PKG_RESTORE:	return "  restore";
     case PKG_VERIFY:	return "   verify";
     case PKG_PRETRANS:	return " pretrans";
     case PKG_POSTTRANS:	return "posttrans";
@@ -829,6 +884,38 @@ static const char * pkgGoalString(pkgGoal goal)
     }
 }
 
+static rpmRC runGoal(rpmpsm psm, pkgGoal goal)
+{
+    rpmRC rc = RPMRC_FAIL;
+    switch (goal) {
+    case PKG_INSTALL:
+	rc = rpmPackageInstall(psm->ts, psm);
+	break;
+    case PKG_ERASE:
+	rc = rpmPackageErase(psm->ts, psm);
+	break;
+    case PKG_RESTORE:
+	rc = rpmPackageRestore(psm->ts, psm);
+	break;
+    case PKG_PRETRANS:
+    case PKG_POSTTRANS:
+    case PKG_VERIFY:
+	rc = runInstScript(psm, goal);
+	break;
+    case PKG_TRANSFILETRIGGERIN:
+	rc = runImmedFileTriggers(psm->ts, psm->te, RPMSENSE_TRIGGERIN,
+				    RPMSCRIPT_TRANSFILETRIGGER, 0);
+	break;
+    case PKG_TRANSFILETRIGGERUN:
+	rc = runImmedFileTriggers(psm->ts, psm->te, RPMSENSE_TRIGGERUN,
+				    RPMSCRIPT_TRANSFILETRIGGER, 0);
+	break;
+    default:
+	break;
+    }
+    return rc;
+}
+
 rpmRC rpmpsmRun(rpmts ts, rpmte te, pkgGoal goal)
 {
     rpmpsm psm = NULL;
@@ -839,40 +926,19 @@ rpmRC rpmpsmRun(rpmts ts, rpmte te, pkgGoal goal)
 	return RPMRC_OK;
 
     psm = rpmpsmNew(ts, te, goal);
+    /* Run pre transaction element hook for all plugins */
     if (rpmChrootIn() == 0) {
-	/* Run pre transaction element hook for all plugins */
 	rc = rpmpluginsCallPsmPre(rpmtsPlugins(ts), te);
+	rpmChrootOut();
+    }
 
-	if (!rc) {
-	    switch (goal) {
-	    case PKG_INSTALL:
-		rc = rpmPackageInstall(ts, psm);
-		break;
-	    case PKG_ERASE:
-		rc = rpmPackageErase(ts, psm);
-		break;
-	    case PKG_PRETRANS:
-	    case PKG_POSTTRANS:
-	    case PKG_VERIFY:
-		rc = runInstScript(psm, goal);
-		break;
-	    case PKG_TRANSFILETRIGGERIN:
-		rc = runImmedFileTriggers(ts, te, RPMSENSE_TRIGGERIN,
-					    RPMSCRIPT_TRANSFILETRIGGER, 0);
-		break;
-	    case PKG_TRANSFILETRIGGERUN:
-		rc = runImmedFileTriggers(ts, te, RPMSENSE_TRIGGERUN,
-					    RPMSCRIPT_TRANSFILETRIGGER, 0);
-		break;
-	    default:
-		break;
-	    }
-	}
-	/* Run post transaction element hook for all plugins */
+    if (!rc)
+	rc = runGoal(psm, goal);
+
+    /* Run post transaction element hook for all plugins (even on failure) */
+    if (rpmChrootIn() == 0) {
 	rpmpluginsCallPsmPost(rpmtsPlugins(ts), te, rc);
-
-	/* XXX an error here would require a full abort */
-	(void) rpmChrootOut();
+	rpmChrootOut();
     }
     rpmpsmFree(psm);
     return rc;
