@@ -52,12 +52,14 @@ static int inhibit(void)
 	dbus_message_unref(reply);
     }
     
-    if (dbus_error_is_set(&err)
-	&& !dbus_error_has_name(&err, DBUS_ERROR_NO_SERVER)
-	&& !dbus_error_has_name(&err, DBUS_ERROR_FILE_NOT_FOUND)) {
-	rpmlog(RPMLOG_WARNING,
+    if (dbus_error_is_set(&err)) {
+	if (!dbus_error_has_name(&err, DBUS_ERROR_NO_SERVER) &&
+	    !dbus_error_has_name(&err, DBUS_ERROR_FILE_NOT_FOUND))
+	{
+	    rpmlog(RPMLOG_WARNING,
 	       "Unable to get systemd shutdown inhibition lock: %s\n",
 		err.message);
+	}
 	dbus_error_free(&err);
     }
 
